@@ -15,8 +15,7 @@ const char *enter_bold_mode = "\x1b[1m";
 const char *enter_dim_mode = "\x1b[2m";
 const char *enter_reverse_mode = "\x1b[7m";
 
-int lines = 79;
-int columns = 24;
+int columns = 80, lines = 24;
 
 int
 setupterm(char *term, int fildes, int *errret)
@@ -25,8 +24,10 @@ setupterm(char *term, int fildes, int *errret)
 
 	if (ioctl(fildes, TIOCGWINSZ, &winsz) == -1)
 		return -1; /* ERR */
-	columns = winsz.ws_col;
-	lines = winsz.ws_row;
+	if (winsz.ws_col > 0)
+		columns = winsz.ws_col;
+	if (winsz.ws_row > 0)
+		lines = winsz.ws_row;
 
 	return 0; /* OK */
 }
